#include <uopt.h>

#include <string>
#include <map>
#include <valarray>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string.h>

// find x
// 0 = rosenbrock(x);

struct MySettings {
};

static double const a = 2, b = 100;

int fsqrt_residual(double const *y_, int64_t N, double *res, int64_t, void *) {

  std::valarray<double> x(y_, N);

  // (a - x[[1]])^2 + b * (x[[2]] - x[[1]]^2)^2

  *res = pow(a - x[0],2) + b * pow(x[1] - x[0]*x[0],2);

  return 0;
}

typedef std::valarray<double> Vector;

int gsqrt_residual(double const *y_, int64_t N, double *res, int64_t, double *gy_, void */*context*/) {

  std::valarray<double> x(y_, N);

  *res = pow(a - x[0],2) + b * pow(x[1] - x[0]*x[0],2);

  // (a - x1)^2 + b * (x2 - x1^2)^2
  // [-2 * (a - x1) - b * 2 * (x2 - x1^2) * 2 * x1, b * 2 * (x2 - x1^2)]

  double const dz1 = -2 * (a - x[0]) - b * 2 * (x[1] - pow(x[0],2)) * 2 * x[0];
  double const dz2 = b * 2 * (x[1] - pow(x[0],2));

  Vector a_x(2);
  a_x[0] = dz1;
  a_x[1] = dz2;

  memcpy(gy_, &a_x[0], sizeof(double)*a_x.size());

  return 0;
}

int Hsqrt_residual(double const *y_, int64_t N, double *res, int64_t, double *gy_, double *Hy_, void */*context*/) {

  std::valarray<double> x(y_, N);

  *res = pow(a - x[0],2) + b * pow(x[1] - x[0]*x[0],2);

  // (a - x1)^2 + b * (x2 - x1^2)^2
  // [-2 * (a - x1) - b * 2 * (x2 - x1^2) * 2 * x1, b * 2 * (x2 - x1^2)]

  double const dz1 = -2 * (a - x[0]) - b * 2 * (x[1] - pow(x[0],2)) * 2 * x[0];
  double const dz2 = b * 2 * (x[1] - pow(x[0],2));

  double const dz1_dz1 = 2 + b * 2 * 2 * x[0] * 2 * x[0] - b * 2 * (x[1] - pow(x[0],2)) * 2;
  double const dz1_dz2 = - b * 2 * 2 * x[0];
  double const dz2_dz1 = -b * 2 * 2 * x[0];
  double const dz2_dz2 = b * 2;

  Vector a_x(N);
  a_x[0] = dz1;
  a_x[1] = dz2;

  Vector H(N*N);
  H[0] = dz1_dz1;
  H[2] = dz1_dz2;

  H[1] = dz2_dz1;
  H[3] = dz2_dz2;

  memcpy(gy_, &a_x[0], sizeof(double)*N);

  memcpy(Hy_, &H[0], sizeof(double)*N*N);

  return 0;
}


int main() {

  std::cout.precision(16);
  std::cerr.precision(16);

  std::valarray<double> x0(1.11, 2);

  long const N = x0.size();

  std::valarray<double> xs(N);

  MySettings context;

  std::map<std::string,std::string> optmap;
  std::ifstream conf("config-rosen.txt", std::ios::in);

  for(long i = 0; conf; ++i) {
    std::string s1, s2;
    getline(conf, s1, '=');
    getline(conf, s2);
    if (s1.empty() || s2.empty()) {
      break;
    }
    optmap[s1] = s2;
  }

  long const nopts = optmap.size();
  char const *options[nopts];
  double ovals[nopts];
  auto it = optmap.begin();
  for (long i = 0; i < nopts; ++i, ++it) {
    options[i] = it->first.c_str();
    ovals[i] = atof(it->second.c_str());
  }

  auto uopt_ouput = [](double const */*x*/, int64_t /*nx*/,
                       double const */*y*/, int64_t /*ny*/,
                       double const */*gf*/, double const */*Hf*/,
                       char const **snames, char const **svals, int64_t nstats,
                       void */*context*/) -> int
                    {
                      std::cout << "UOPT: ";
                      for (long i = 0; i < nstats; ++i) {
                        if (i > 0) std::cout << ", ";
                        std::cout << snames[i] << "=" << svals[i];
                      }
                      std::cout << "\n";
                      return 0;
                    };

  double ys = 0;
  int rc = uopt(&x0[0], &xs[0], x0.size(), &ys, 1,
                fsqrt_residual, gsqrt_residual, 0, Hsqrt_residual, 0,
                uopt_ouput,
                options, ovals, nopts,
                &context);

  std::cerr << "result: " << rc << "\n";

  std::cout << "sol: \n";
  for (unsigned i = 0; i < xs.size(); ++i) {
    if (i > 0) {
      std::cout << ", ";
    }
    std::cout << xs[i];
  }
  std::cout << "\n";

}
