#include <uopt.h>

#include <mex.h>

#include <cstring>
#include <cstdio>
#include <vector>
#include <map>
#include <string>

#define mxIsScalar(xk) (mxGetNumberOfElements(xk) == 1)

void mexFunction(int nlhs, mxArray *plhs[], int nrhs,  const mxArray *prhs[]) {

  char const *funcName = mexFunctionName();

  if (nrhs != 7) {
    mexErrMsgIdAndTxt("uopt:error:nargs", "The function %s must receive 7 arguments, got %d", funcName, nrhs);
  }

  mxArray const *xin      = prhs[0];
  mxArray const *fHandle  = prhs[1];
  mxArray const *gHandle  = prhs[2];
  mxArray const *HHandle  = prhs[3];
  mxArray const *HvHandle = prhs[4];
  mxArray const *oHandle  = prhs[5];
  mxArray const *options  = prhs[6];

  if (mxGetClassID(prhs[0]) != mxDOUBLE_CLASS) {
    mexErrMsgIdAndTxt("uopt:arg:type", "The first argument of %s must be a double array", funcName);
  }

  if (mxGetClassID(prhs[1]) != mxFUNCTION_CLASS) {
    mexErrMsgIdAndTxt("uopt:arg:type", "The second argument of %s must be a function handle", funcName);
  }
  if (mxGetClassID(prhs[2]) != mxFUNCTION_CLASS) {
    mexErrMsgIdAndTxt("uopt:arg:type", "The third argument of %s must be a function handle", funcName);
  }
  if (mxGetClassID(prhs[3]) != mxFUNCTION_CLASS
      and (mxGetClassID(prhs[3]) != mxDOUBLE_CLASS or !mxIsEmpty(prhs[3]))) {
    mexErrMsgIdAndTxt("uopt:arg:type", "The 4th argument of %s must be a function handle or empty", funcName);
  }
  if (mxGetClassID(prhs[4]) != mxFUNCTION_CLASS
      and (mxGetClassID(prhs[4]) != mxDOUBLE_CLASS or !mxIsEmpty(prhs[4]))) {
    mexErrMsgIdAndTxt("uopt:arg:type", "The 5th argument of %s must be a function handle or empty", funcName);
  }
  if (mxGetClassID(prhs[5]) != mxFUNCTION_CLASS) {
    mexErrMsgIdAndTxt("uopt:arg:type", "The 6th argument of %s must be a function handle", funcName);
  }

  if (mxGetClassID(prhs[6]) != mxSTRUCT_CLASS) {
    mexErrMsgIdAndTxt("uopt:arg:type", "The 7th argument of %s must be a struct", funcName);
  }
  if (mxGetNumberOfElements(prhs[6]) != 1) {
    mexErrMsgIdAndTxt("uopt:arg:type", "The 7th argument of %s must be a scalar struct", funcName);
  }


  auto fhClosure = [&fHandle](double const *xData, int64_t n, double *yData, int64_t /*ny*/) -> int {

    mxArray *x = mxCreateDoubleMatrix(n, 1, mxREAL);
    memcpy(mxGetPr(x), xData, sizeof(double)*n);

    mxArray *result[1];
    mxArray *args[2] = {(mxArray*) fHandle, x};
    int rc = mexCallMATLAB(1, result, 2, args, "feval");
    if (rc != 0) {
      mexErrMsgIdAndTxt("uopt:fcall", "Callback to MATLAB function failed");
    }
    mxArray * xk = result[0];

    if (xk == 0) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function f must return one result");
    }

    if (!mxIsDouble(xk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function f must return a real value");
    }
    if (!mxIsScalar(xk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function f must return a scalar value");
    }

    double *xkData = mxGetPr(xk);
    *yData = xkData[0];

    return 0;
  };


  auto ghClosure = [&gHandle, &funcName](double const *xData, int64_t n, double *y, int64_t /*ny*/, double *g) -> int {

    mxArray *x = mxCreateDoubleMatrix(n, 1, mxREAL);
    memcpy(mxGetPr(x), xData, sizeof(double)*n);

    mxArray *result[2];
    mxArray *args[2] = {(mxArray*) gHandle, x};
    int rc = mexCallMATLAB(2, result, 2, args, "feval");
    if (rc != 0) {
      mexErrMsgIdAndTxt("uopt:fcall", "Callback to MATLAB function failed");
    }

    mxArray * xk = result[0];
    mxArray * gk = result[1];

    if (xk == 0 || gk == 0) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function g must return two results");
    }

    if (!mxIsDouble(xk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function g must return a real value");
    }
    if (!mxIsScalar(xk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function g must return a scalar value");
    }

    if (!mxIsDouble(gk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function g must return a real vector");
    }
    if (int64_t(mxGetNumberOfElements(gk)) != n) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function g must return a real vector of same length as x, but %d != %d", n, mxGetNumberOfElements(gk));
    }

    double *xkData = mxGetPr(xk);
    *y = xkData[0];

    double *gkData = mxGetPr(gk);
    memcpy(g, gkData, sizeof(double)*n);

    return 0;
  };


  auto HhClosure = [&HHandle](double const *xData, int64_t n, double *y, int64_t /*ny*/, double *g, double *H) -> int {

    mxArray *x = mxCreateDoubleMatrix(n, 1, mxREAL);
    memcpy(mxGetPr(x), xData, sizeof(double)*n);

    mxArray *result[3];
    mxArray *args[2] = {(mxArray*) HHandle, x};
    int rc = mexCallMATLAB(3, result, 2, args, "feval");
    if (rc != 0) {
      mexErrMsgIdAndTxt("uopt:fcall", "Callback to MATLAB function failed");
    }

    mxArray * xk = result[0];
    mxArray * gk = result[1];
    mxArray * Hk = result[2];

    if (xk == 0 || gk == 0 || Hk == 0) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function H must return three results");
    }

    if (!mxIsDouble(xk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function H must return a real value");
    }
    if (!mxIsScalar(xk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function H must return a scalar value");
    }

    if (!mxIsDouble(gk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function H must return a real vector");
    }
    if (int64_t(mxGetNumberOfElements(gk)) != n) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function H must return a real vector of same length as x, but %d != %d", n, mxGetNumberOfElements(gk));
    }

    if (!mxIsDouble(Hk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function H must return a real vector");
    }
    if (int64_t(mxGetNumberOfElements(Hk)) != n*n) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function H must return a real vector of same length as x, but %d != %d", n*n, mxGetNumberOfElements(Hk));
    }

    double *xkData = mxGetPr(xk);
    *y = xkData[0];

    double *gkData = mxGetPr(gk);
    memcpy(g, gkData, sizeof(double)*n);

    double *HkData = mxGetPr(Hk);
    memcpy(H, HkData, sizeof(double)*n*n);

    return 0;
  };


  auto HhvClosure = [&HvHandle](double const *xData, int64_t nx, double *y, int64_t /*ny*/,
                                double const *dxData, int64_t ndx, double *gf, double *Hfv) -> int {

    mxArray *x = mxCreateDoubleMatrix(nx, 1, mxREAL);
    memcpy(mxGetPr(x), xData, sizeof(double)*nx);

    mxArray *dx = mxCreateDoubleMatrix(nx, ndx, mxREAL);
    memcpy(mxGetPr(dx), dxData, sizeof(double)*nx*ndx);

    mxArray *result[3];
    mxArray *args[3] = {(mxArray*) HvHandle, x, dx};
    int rc = mexCallMATLAB(3, result, 3, args, "feval");
    if (rc != 0) {
      mexErrMsgIdAndTxt("uopt:fcall", "Callback to MATLAB function failed");
    }

    mxArray * xk = result[0];
    mxArray * gk = result[1];
    mxArray * Hk = result[2];

    if (xk == 0 || gk == 0 || Hk == 0) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function Hv must return three results");
    }

    if (!mxIsDouble(xk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function Hv must return a real value");
    }
    if (!mxIsScalar(xk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function Hv must return a scalar value");
    }

    if (!mxIsDouble(gk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function Hv must return a real vector");
    }
    if (int64_t(mxGetNumberOfElements(gk)) != nx) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function Hv must return a real vector of same length as x, but %d != %d", nx, mxGetNumberOfElements(gk));
    }

    if (!mxIsDouble(Hk)) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function Hv must return a real vector");
    }
    if (int64_t(mxGetNumberOfElements(Hk)) != nx*ndx) {
      mexErrMsgIdAndTxt("uopt:fcall:result", "Function Hv must return a real matrix of size [%d,%d], but numel is %d != %d",
                        nx, ndx, nx*ndx, mxGetNumberOfElements(Hk));
    }

    double *xkData = mxGetPr(xk);
    *y = xkData[0];

    double *gkData = mxGetPr(gk);
    memcpy(gf, gkData, sizeof(double)*nx);

    double *HkData = mxGetPr(Hk);
    memcpy(Hfv, HkData, sizeof(double)*nx*ndx);

    return 0;
  };


  auto ohClosure = [&oHandle](double const *xData, int64_t n, double const *yData, int64_t ny,
                              char const **onames, char const **ovals, int64_t const no) -> int {

    int oreturncode = 0;

    mxArray *x = mxCreateDoubleMatrix(n, 1, mxREAL);
    memcpy(mxGetPr(x), xData, sizeof(double)*n);

    mxArray *y = mxCreateDoubleMatrix(ny, 1, mxREAL);
    memcpy(mxGetPr(y), yData, sizeof(double)*ny);

    mxArray *status = mxCreateStructMatrix(1l, 1l, no, onames);
    for (int64_t i = 0; i < no; ++i) {
      mxArray *fvalue = mxCreateString(ovals[i]);
      mxSetFieldByNumber(status, 0, i, fvalue);
    }

    mxArray *result[1];
    mxArray *args[4] = {(mxArray*) oHandle, x, y, status};
    int rc = mexCallMATLAB(1, result, 4, args, "feval");
    if (rc != 0) {
      mexErrMsgIdAndTxt("uopt:fcall", "Callback to MATLAB function failed");
    }

    mxArray * sk = result[0];

    if (sk) {
      if (!mxIsDouble(sk)) {
        mexErrMsgIdAndTxt("uopt:fcall:result", "Function output must return a real value");
      }
      if (!mxIsScalar(sk)) {
        mexErrMsgIdAndTxt("uopt:fcall:result", "Function output must return a scalar value");
      }
      oreturncode = mxGetPr(sk)[0];
    }

    return oreturncode;
  };


  struct CallInfo {
    void *f1, *g1, *h1, *hv, *o1;
  } mycallInfo;
  mycallInfo.f1 = &fhClosure;
  mycallInfo.g1 = &ghClosure;
  mycallInfo.h1 = &HhClosure;
  mycallInfo.hv = &HhvClosure;
  mycallInfo.o1 = &ohClosure;

  fhandle fh = [](double const *x, int64_t n, double *y, int64_t ny, void *arg) -> int {
        CallInfo *ci = static_cast<CallInfo*>(arg);
        auto fhClosLoc = static_cast<decltype(fhClosure)*>(ci->f1);
        (*fhClosLoc)(x, n, y, ny);
        return 0;
  };

  Jhandle gh = [](double const *x, int64_t nx, double *y, int64_t ny, double *Jf, void *arg) -> int {
        CallInfo *ci = static_cast<CallInfo*>(arg);
        auto ghClosLoc = static_cast<decltype(ghClosure)*>(ci->g1);
        (*ghClosLoc)(x, nx, y, ny, Jf);
        return 0;
  };

  Hhandle Hh = [](double const *x, int64_t nx, double *y, int64_t ny,
                  double *gf, double *Hf, void *arg) -> int {
        CallInfo *ci = static_cast<CallInfo*>(arg);
        auto HhClosLoc = static_cast<decltype(HhClosure)*>(ci->h1);
        (*HhClosLoc)(x, nx, y, ny, gf, Hf);
        return 0;
  };
  if (mxIsEmpty(HHandle)) {
    Hh = 0;
  }

  Hvhandle Hhv = [](double const *x, int64_t nx, double *y, int64_t ny,
                    double const *dx, int64_t ndx, double *gf, double *Hfv, void *arg) -> int {
        CallInfo *ci = static_cast<CallInfo*>(arg);
        auto HhvClosLoc = static_cast<decltype(HhvClosure)*>(ci->hv);
        (*HhvClosLoc)(x, nx, y, ny, dx, ndx, gf, Hfv);
        return 0;
  };
  if (mxIsEmpty(HvHandle)) {
    Hhv = 0;
  }

  shandle oh = [](double const *x, int64_t nx, double const *y, int64_t ny,
                  double const *, double const *,
                  char const **snames, char const **svals, int64_t nstats,
                  void *arg) -> int {
        CallInfo *ci = static_cast<CallInfo*>(arg);
        auto ohClosLoc = static_cast<decltype(ohClosure)*>(ci->o1);
        int rc = (*ohClosLoc)(x, nx, y, ny, snames, svals, nstats);
        return rc;
  };

  size_t const n = mxGetNumberOfElements(xin);
  size_t const ny = 1;
  double *xData = mxGetPr(xin);

  mxArray *xs = mxCreateDoubleMatrix(n, 1, mxREAL);
  double *xsData = mxGetPr(xs);

  mxArray *ys = mxCreateDoubleMatrix(ny, 1, mxREAL);
  double *ysData = mxGetPr(ys);

  std::map<std::string, double> optsMap;
  std::vector<char const*> optNames;
  std::vector<double> optValues;

  int const nfields = mxGetNumberOfFields(options);
  for (int i = 0; i < nfields; ++i) {
    char const * fname = mxGetFieldNameByNumber(options, i);
    mxArray const *felem = mxGetFieldByNumber(options, 0, i);
    if (!mxIsDouble(felem)) {
      mexErrMsgIdAndTxt("uopt:option:type", "Option value %s must be a real number", fname);
    }
    if (!mxIsScalar(felem)) {
      mexErrMsgIdAndTxt("uopt:option:type", "Option value %s must be a real scalar", fname);
    }
    optNames.push_back(fname);
    optValues.push_back(mxGetPr(felem)[0]);
    optsMap[fname] = mxGetPr(felem)[0];
  }

  if (optsMap.find("enableExit") == optsMap.end()) {
    optNames.push_back("enableExit");
    optValues.push_back(0);
  }
  if (optsMap.find("enableSignalHandler") == optsMap.end()) {
    optNames.push_back("enableSignalHandler");
    optValues.push_back(0);
  }
  if (optsMap.find("enableFloatExcept") == optsMap.end()) {
    optNames.push_back("enableFloatExcept");
    optValues.push_back(0);
  }

  int mrc = uopt(xData, xsData, n, ysData, ny,
                 fh, gh, 0, Hh, Hhv, oh,
                 &optNames[0], &optValues[0], optValues.size(), &mycallInfo);

  plhs[0] = xs;
  if (nlhs > 1) {
    plhs[1] = ys;
  }
  if (nlhs > 2) {
    plhs[2] = mxCreateDoubleScalar(mrc);
  }

}
