import uopt
import numpy as np

import sphere3d

import random
import time
import json

def run(testcase=None, opts=None):

    if testcase:
        with open('%s/srand.txt' % (testcase['name'],), 'r') as f:
            srand = float(f.read())
    else:
        srand = time.time()
        with open('srand.txt', 'w') as f:
            f.write('%.17g' % srand)
    random.seed(srand)

    obj = sphere3d.Sphere3DDataset(30)
    obj.opts = {k: opts[k] for k in opts}
    if 'checkres' in opts:
        del opts['checkres']

    if testcase:
        with open('%s/options.json' % (testcase['name'],), 'r') as f:
            opts = json.loads(f.read())
        for k in testcase['opts']:
            opts[k] = testcase['opts'][k]
    else:
        if opts is None:
            opts = {'sdirs': 8,
                    'maxit': 100,
                    }
        with open('options.json', 'w') as f:
            f.write(json.dumps(opts))

    R = random.random()
    P = [random.random(), random.random(), random.random()]

    obj.initSamples(R, P)

    R0 = random.random()
    if 'x.initrand' in opts:
        P0 = [random.random(), random.random(), random.random()]
    else:
        P0 = list(np.mean(obj.samples, axis=0))

    fvalinit = obj.dists(R0, P0)

    #print('fs', obj.dists(R, P))
    #print('f0', fvalinit)

    x0 = np.ndarray((4,))
    x0[:] = [R0] + P0
    xs = np.ndarray((4,))

    fvals = np.ndarray((obj.M,))
    #sphere3d.fhandle(x0, fvals, obj)
    #print('f0', fvals)

    #Jac = np.ndarray((obj.M,4))
    #sphere3d.Jhandle(x0, fvals, Jac, obj)
    #print('J0', Jac)

    (rc, info) = uopt.usolve(x0, xs, fvals, sphere3d.fhandle, sphere3d.Jhandle, sphere3d.Jvhandle, None, sphere3d.ohandle, opts, obj)

    print('usolve status code:', rc)
    print('usolve xs:', xs)
    print('usolve fs:', np.linalg.norm(fvals))
    xt = np.ndarray((4,))
    xt[:] = [R] + P
    print('expected:', xt)
    print('deviation:', xt - xs, np.linalg.norm(xt - xs))

    if rc == 0:
        Rs, Ps = sphere3d.unpack(xs)
        print('Solution: found sphere with R=%g at %s' %(Rs, Ps))
        print('Dev. in radius: Rs - R = %g' %(Rs-R,))
        print('Dev. in position: Ps - P = %g' %(np.linalg.norm(Ps-P),))

    if len(obj.errors):
        print('Errors occured')
        for i in obj.errors:
            print(i)
        exit(2)

    return rc

def rerun(testcase):
    run(testcase)

if __name__ == "__main__":

    common_opts = {'maxit': 100}

    test_opts = [
        {'tolObjRel': -1,
         'tolEnhRel': -1,
         'tolStepRel': -1,
         'tolGradRel': -1,
         'tolObjAbs': -1,
         'tolEnhAbs': -1,
         'tolStepAbs': -1,
         'tolGradAbs': -1,
         },
        {'tolObjRel': -1,
         'tolEnhRel': -1,
         'tolStepRel': -1,
         'tolGradRel': -1,
         'tolObjAbs': 1e-6,
         'tolEnhAbs': -1,
         'tolStepAbs': -1,
         'tolGradAbs': -1,
         'checkres': lambda s: s['exit'] == 'tolObjAbs reached'
         },
        {'tolObjRel': -1,
         'tolEnhRel': -1,
         'tolStepRel': -1,
         'tolGradRel': -1,
         'tolObjAbs': -1,
         'tolEnhAbs': 1e-6,
         'tolStepAbs': -1,
         'tolGradAbs': -1,
         'checkres': lambda s: s['exit'] == 'tolEnhAbs reached'
         },
        {'tolObjRel': -1,
         'tolEnhRel': -1,
         'tolStepRel': -1,
         'tolGradRel': -1,
         'tolObjAbs': -1,
         'tolEnhAbs': -1,
         'tolStepAbs': 1e-6,
         'tolGradAbs': -1,
         'checkres': lambda s: s['exit'] == 'tolStepAbs reached'
         },
        {'tolObjRel': -1,
         'tolEnhRel': -1,
         'tolStepRel': -1,
         'tolGradRel': -1,
         'tolObjAbs': -1,
         'tolEnhAbs': -1,
         'tolStepAbs': -1,
         'tolGradAbs': 10,
         'checkres': lambda s: s['exit'] == 'tolGradAbs reached'
         },
        {'tolObjRel': 1e-5,
         'tolEnhRel': -1,
         'tolStepRel': -1,
         'tolGradRel': -1,
         'tolObjAbs': -1,
         'tolEnhAbs': -1,
         'tolStepAbs': -1,
         'tolGradAbs': -1,
         'checkres': lambda s: s['exit'] == 'tolObjRel reached'
         },
        {'tolObjRel': -1,
         'tolEnhRel': 1e-5,
         'tolStepRel': -1,
         'tolGradRel': -1,
         'tolObjAbs': -1,
         'tolEnhAbs': -1,
         'tolStepAbs': -1,
         'tolGradAbs': -1,
         'checkres': lambda s: s['exit'] == 'tolEnhRel reached'
         },
        {'tolObjRel': -1,
         'tolEnhRel': -1,
         'tolStepRel': 1e-5,
         'tolGradRel': -1,
         'tolObjAbs': -1,
         'tolEnhAbs': -1,
         'tolStepAbs': -1,
         'tolGradAbs': -1,
         'checkres': lambda s: s['exit'] == 'tolStepRel reached'
         },
        {'tolObjRel': -1,
         'tolEnhRel': -1,
         'tolStepRel': -1,
         'tolGradRel': 1e-6,
         'tolObjAbs': -1,
         'tolEnhAbs': -1,
         'tolStepAbs': -1,
         'tolGradAbs': -1,
         'checkres': lambda s: s['exit'] == 'tolGradRel reached'
         }
    ]

    for (i,o) in enumerate(test_opts):
        for co in common_opts:
            test_opts[i][co] = common_opts[co]
        rc = run(opts=test_opts[i])
        if rc != 0:
            break
    exit(rc)
