"""UOpt Python

The UOpt python module provides and easy interface to the UOpt shared library in Python.

"""

from ctypes import CDLL, POINTER, CFUNCTYPE, c_int, c_long
from ctypes import c_longlong, c_double, c_void_p, c_char, c_char_p
from ctypes import sizeof, util

import numpy as np
import os
import platform

if sizeof(c_long) == 8:
    c_int64 = c_long
else:
    c_int64 = c_longlong

DPTR = POINTER(c_double)
CPTRPTR = POINTER(c_char_p)

FFUNC = CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, c_void_p)
JFUNC = CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, c_void_p)
JVFUNC= CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, c_int64, DPTR, c_void_p)
OFUNC = CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, DPTR,
                  CPTRPTR, CPTRPTR, c_int64, c_void_p)

def find_uopt():
    """Find the path to the UOpt shared library."""
    path = util.find_library("uopt")
    if path is None:
        if os.name == 'nt':
            path = 'uopt.dll'
        else:
            path = 'libuopt.so'
    return path

def fix_opts(opts):
    """Set certain options in opts when they are unset."""
    if not 'enableExit' in opts:
        opts['enableExit'] = 0
    if not 'enableSignalHandler' in opts:
        opts['enableSignalHandler'] = 0
    if not 'enableFloatExcept' in opts:
        opts['enableFloatExcept'] = 0
    return opts


def statusDefault(x, y, status, udata):
    """Default status function, prints a line with key=value fields for each status item."""
    maxLen = max(14, max([len(f)+1+len(status[f]) for f in status]))
    joinf = ""
    if maxLen > 20:
        joinf = "\n"
    print(joinf.join([("%%- %ds" % (maxLen+2,)) %
                       ("%s=%s" % (f, status[f])) for f in status]))


def statusHist(chain=statusDefault):
    """Returns a status function.

    Returns a status to be use with the parameter s of :py:func:`uopt`
    and :py:func:`usolve`. After running, this status function can be
    given to :py:func:`getEntries`, :py:func:`getSeries` and
    :py:func:`mkPlot`

    Parameters
    ----------
    chain : function

        If given, call this function as well whenever the status
        function is called. Defaults to statusDefault.

    Returns
    -------
    function
        Status function to be used with usolve or uopt.

    """
    hist = []
    def inner(*args):
        # args == x, y, stats, udata
        nonlocal hist
        if isinstance(args[0], str):
            if args[0] == "get":
                return hist
            elif args[0] == "clear":
                hist = []
        else:
            hist += [args[2]]
        if chain:
            chain(*args)
    return inner


def getEntries(sfun, snames):
    """Returns matching fields from the history of the status function.

    Returns fields from the history of the status function that have
    all of the fields listed in snames.

    Parameters
    ----------
    sfun : function
        Status function.

    snames : list of strings
        Fields required.

    Returns
    -------
    list
        Matching history items.

    """
    hist = sfun("get")
    return [hist[i] for i in range(len(hist)) if all([s in hist[i] for s in snames])]


def getSeries(sfun, sname):
    """Returns matching key values of the field sname from the history
    of the status function.

    Parameters
    ----------
    sfun : function
        Status function.

    sname : str
        Desired fieldname.

    Returns
    -------
    list
        Matching key values as float.

    """
    entries = getEntries(sfun, sname)
    return [float(entries[sname]) for i in getEntries(sfun, sname) ]


def mkPlot(sfun, field, over='it',
           show=False, save=None, outdir='.',
           title=None, subtitle=None,
           xunit=None, yunit=None,
           xlim=None, ylim=None,
           xscale='linear', yscale='log',
           usetex=False, hardwareinfo=False):
    """Generate a plot of the iteration history field over the other field over.

    Generate a plot of the iteration history with :py:mod:`matplotlib`, plotting
    the values of history field "field" over the other field "over".

    When save is a string the figure is saved under that filename.
    Then show is True, the plot is shown in the foreground.

    Parameters
    ----------
    sfun : function
        Status function.

    field : str
        Desired fieldname for y axsis, like "J" or "Er".

    over : str
        Desired fieldname for x axsis, like "it" or "tj".

    title : str
        Plot title. When None, a title is automatically generated.

    subtitle : str
        Plot subtitle. When set to 'pyinfo', a subtitle with python
        information is automatically generated.

    show : bool
        Show plot.

    save : str
        Save plot as file save

    outdir : str
        Save plot in director outdir.

    xunit : str
        Add string " (xunit)" to xlabel.

    yunit : str
        Add string " (yunit)" to ylabel.

    xlim : tuple
        Set axis xlim.

    ylim : tuple
        Set axis ylim.

    xscale : str
        Set axis xscale.

    yscale : str
        Set axis yscale.

    usetex : bool
        Set usetex in matplotlib.

    hardwareinfo : bool
        Add more lines to the plot title describing the hardware and OS.

    Returns
    -------
    figure, axis
        Returns tuple of figure and axis handles.

    """

    if subtitle == "pyinfo":
        subtitle = f"With\
 {platform.python_implementation()} {platform.python_compiler()} OS {platform.system()} {platform.release()}\
 {' '.join(platform.libc_ver())}"
    if hardwareinfo:
        import cpuinfo
        info = cpuinfo.get_cpu_info()
        toKB = lambda x: f'{int(x/(2**10))}k' if x<(2**20) else f'{int(x/(2**20))}M'
        cacheinfo = f'L1(I/D)/L2/L3 = \
({toKB(info["l1_instruction_cache_size"])}/{toKB(info["l1_data_cache_size"])})/\
{toKB(info["l2_cache_size"])}/\
{toKB(info["l3_cache_size"])}'
        if usetex:
            hwinfo = f'{info["brand_raw"]}, {cacheinfo}'
        else:
            hwinfo = f'{info["brand_raw"]}\n{cacheinfo}'

    if title is None:
        title = f'{field} over {over}'

    if usetex:
        title = '\\begin{center}\\textbf{' + title + '}'
        if subtitle:
            title += '\\\\ {\\textrm ' + subtitle + '}'
        if hardwareinfo:
            title += '\\\\ {\\small ' + hwinfo + '}'
        title += '\\end{center}'
        ntitlelns = len(title.split('\\\\'))
    else:
        if subtitle:
            title += '\n' + subtitle
        if hardwareinfo:
            title += '\n' + hwinfo
        ntitlelns = len(title.split('\n'))


    entries = getEntries(sfun, [field, over])
    xs = [float(e[over]) for e in entries]
    ys = [float(e[field]) for e in entries]

    if len(xs) == 0:
        return

    xlabel = f'{over}' + (f' ({xunit})' if xunit is not None else '')
    ylabel = f'{field}' + (f' ({yunit})' if yunit is not None else '')

    try:
        import matplotlib.pyplot as plt
    except:
        print('Failed to import matplotlib')
        return None, None

    if usetex:
        plt.rc('text', usetex=True)
    else:
        plt.rc('text', usetex=False)
    plt.rc('font', family='serif')

    fig, ax = plt.subplots()

    fig.tight_layout(rect=[0.1,0.05,.9,1 - 0.05 * ntitlelns])

    ax.plot(xs, ys, linewidth=2.0)
    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    ax.set_xlabel(f'{over}' + (f' ({xunit})' if xunit is not None else ''))
    ax.set_ylabel(f'{field}' + (f' ({yunit})' if yunit is not None else ''))
    ax.set_title(title)
    if xlim is not None:
        ax.set_xlim(*xlim)
    if ylim is not None:
        ax.set_ylim(*ylim)

    plt.draw()

    if save is None and not show:
        save = os.path.join(outdir, f'plot_{field}_over_{over}.png')
        print(f'Saving plot to file {save}')

    if save:
        plt.savefig(save)

    if show:
        plt.show()

    return fig, ax


def uopt(x0, xs, ys, f, g, H=None, Hv=None, Jv=None, s=statusDefault, opts={}, udata=None):
    """Run the UOpt nonlinear optimizer.

    Run the UOpt nonlinear optimizer to minimize f(x0).

    When H or Hv are provided, the Newton search directions are
    enabled. Otherwise only CG search directions are used.

    Parameters
    ----------
    x0 : N-array
        Starting values.

    xs : N-array
        Optimized values.

    ys : M-array
        Solution value. Must have size M=1.

    f : function(x, y, udata)
        Optimization objective function. Must write a scalar value
        into y as the result.

    g : function(x, y, dxdy, udata)
        Function that computes the derivative of the objective
        function. Must write the scalar result into y and the
        derivative into dxdy.

    H : function(x, y, dxdy, d2xdy2, udata)
        Function that computes the Hessian of the objective
        function. Must write the scalar result into y, the
        derivative into dxdy and the Hessian into d2xdy2.

    Hv : function(x, y, dx, dxdy, d2xdy2, udata)
        Function that computes Hessian-vector products of the
        objective function. The derivative direction is in dx and has
        shape (N, NDD). Hv must write the scalar result into y, the
        (full) gradient into dxdy and the NDD Hessian-vector products
        into d2xdy2.

    Jv : function(x, y, dx, dxdy, udata)
        Function that computes gradient-vector dot products of the
        objective function. The derivative direction is in dx and has
        shape (N, NDD). Jv must write the scalar result into y and the
        NDD gradient-vector dot products into dxdy.

    s : function(x, y, status, udata)
        Function that is called for status output. It is given the
        current state x, the current result y, a dictionary with
        status values and the user data udata. The function s should
        return None or zero (0) to continue, or a non-zero value to
        stop the iteration.

    opts : dict
        Parameters to uopt, like tolObjAbs, tolObjRel, sdirs, maxiter, etc..

    udata : any
        User data that is passed to all callbacks.

    Returns
    -------
    int
        Returns a status code. Zero indicates success.

    """

    return uopt0(x0, xs, ys, f, g, H, Hv, Jv, s, opts, udata)

def uopt0(x0, xs, ys, fhandle, ghandle, Hhandle, Hvhandle, Jvhandle, ohandle, opts, udata):

    mkarr = np.ctypeslib.as_array

    def fstub(x, nx, y, ny, userdata):
        fhandle(mkarr(x,(nx,)), mkarr(y,(ny,)), udata)
        return 0

    def gstub(x, n, y, ny, g, userdata):
        ghandle(mkarr(x,(n,)), mkarr(y,(ny,)), mkarr(g,(n,)), udata)
        return 0

    def Jacvstub(x, nx, y, ny, dx, ndx, Jv, userdata):
        if ndx > 1:
            raise("not implemented")
        Jvhandle(mkarr(x,(nx,)), mkarr(y,(ny,)), mkarr(dx,(nx,ndx)), mkarr(Jv,(ny,ndx)), udata)
        return 0

    def Hstub(x, n, y, ny, g, H, userdata):
        Hhandle(mkarr(x,(n,)), mkarr(y,(ny,)), mkarr(g,(n,)), mkarr(H,(n,n)), udata)
        return 0

    def Hvstub(x, n, y, ny, dx, ndx, g, H, userdata):
        Hvhandle(mkarr(x,(n,)), mkarr(y,(ny,)), mkarr(dx,(n,ndx)), mkarr(g,(n,)), mkarr(H,(n,ndx)), udata)
        return 0

    def ostub(x, nx, y, ny, g, H, onames, ovals, no, userdata):
        if ohandle is None:
            return 0
        d = {onames[i].decode('utf8'): ovals[i].decode('utf8') for i in range(no)}
        r = ohandle(mkarr(x,(nx,)), mkarr(y,(ny,)), d, udata)
        return int(r) if r is not None else 0

    path = find_uopt()
    try:
        dll = CDLL(path)
    except BaseException as E:
        print("Failed to find libuopt", E)
        raise(E)
    cuopt = dll.uopt

    opts = fix_opts(opts)

    names  = [bytes(k, 'utf-8') for k in opts.keys()]
    values = [float(k) for k in opts.values()]

    HFUNC = CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, DPTR, c_void_p)
    HVFUNC= CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, c_int64, DPTR, DPTR, c_void_p)

    cuopt.return_type = c_int
    cuopt.argtypes = (DPTR, DPTR, c_int64, DPTR, c_int64,
                      FFUNC, JFUNC, JVFUNC, HFUNC, HVFUNC, OFUNC,
                      CPTRPTR, DPTR, c_int64, c_void_p)
    cuopt.errcheck = my_errcheck

    ffunc  = FFUNC(fstub)
    gfunc  = JFUNC(gstub)
    Jvfunc = JVFUNC(Jacvstub if Jvhandle is not None else 0)
    if Hhandle is None:
        Hfunc = HFUNC(0)
    else:
        Hfunc  = HFUNC(Hstub)
    if Hvhandle is None:
        Hvfunc = HVFUNC(0)
    else:
        Hvfunc = HVFUNC(Hvstub)
    ofunc  = OFUNC(ostub)

    cx0 = np.ctypeslib.as_ctypes(x0)
    cxs = np.ctypeslib.as_ctypes(xs)
    cys = np.ctypeslib.as_ctypes(ys)
    STRARR = c_char_p * len(names)
    DBLARR = c_double * len(names)
    cnames = STRARR()
    cvalues = DBLARR()
    cnames[:] = names
    cvalues[:] = values
    res = cuopt(cx0, cxs, x0.size, cys, ys.size,
                ffunc, gfunc, Jvfunc, Hfunc, Hvfunc, ofunc,
                cnames, cvalues, len(names), None)
    return res

def usolve(x0, xs, ys, f, J, Jv=None, JTv=None, s=statusDefault, opts={}, udata=None):
    """Run the UOpt nonlinear solver.

    Run the UOpt nonlinear solver to find f(xs) = ys = 0.

    When Jv and JTv are provided to compute forward and reverse
    directional derivatives, the matrix-free method is used, otherwise
    J must be provided for the full derivative.

    Parameters
    ----------
    x0 : N-array
        Starting values.

    xs : N-array
        Optimized values.

    ys : M-array
        Solution values.

    f : function(x, y, udata)
        Optimization objective function. Must write a scalar value
        into y as the result.

    J : function(x, y, dxdy, udata)
        Function that computes the derivative of the objective
        function. Must write the scalar result into y and the
        derivative into dxdy.

    Jv : function(x, y, dx, dxdy, udata)
        Function that computes directional derivatives of the
        objective function. Must write the scalar result into y, and
        the directional derivative d/dt f(x+t*dx) evaluated at t=0
        into dxdy.

    JTv : function(x, y, dy, dxdy, udata)
        Function that computes reverse directional derivatives of the
        objective function. Must write the scalar result into y, and
        the derivative d/dx dot(f(x), dy) into dxdy.

    s : function(x, y, status, udata)
        Function that is called for status output. It is given the
        current state x, the current result y, a dictionary with
        status values and the user data udata. The function s should
        return None or zero (0) to continue, or a non-zero value to
        stop the iteration.

    opts : dict
        Parameters to uopt, like tolObjAbs, tolObjRel, maxiter, etc..

    udata : any
        User data that is passed to all callbacks.

    Returns
    -------
    int
        Returns a status code. Zero indicates success.

    """
    return usolve0(x0, xs, ys, f, J, Jv, JTv, s, opts, udata)

def usolve0(x0, xs, ys, fhandle, Jhandle, Jvhandle, JTvhandle, ohandle, opts, udata):

    mkarr = np.ctypeslib.as_array
    runinfo = {'status': {}}

    def fstub(x, nx, y, ny, userdata):
        fhandle(mkarr(x, (nx,)), mkarr(y, (ny,)), udata)
        return 0

    def Jacstub(x, nx, y, ny, Jac, userdata):
        Jarr = mkarr(Jac, (ny, nx))
        Jarr = np.lib.stride_tricks.as_strided(Jarr, strides=(sizeof(c_double), ny*sizeof(c_double)))
        Jhandle(mkarr(x, (nx,)), mkarr(y, (ny,)), Jarr, udata)
        return 0

    def Jacvstub(x, nx, y, ny, dx, ndx, Jv, userdata):
        if ndx > 1:
            raise("not implemented")
        Jvhandle(mkarr(x, (nx,)), mkarr(y, (ny,)), mkarr(dx, (nx,ndx)), mkarr(Jv, (ny,ndx)), udata)
        return 0

    def JacTvstub(x, nx, y, ny, dy, ndy, JTv, userdata):
        if ndy > 1:
            raise("not implemented")
        JTvhandle(mkarr(x, (nx,)), mkarr(y, (ny,)), mkarr(dy, (ndy, ny)), mkarr(JTv, (ndy, nx)), udata)
        return 0

    def ostub(x, nx, y, ny, J, _, onames, ovals, no, userdata):
        if ohandle is None:
            return 0
        d = {onames[i].decode('utf8'): ovals[i].decode('utf8') for i in range(no)}
        r = ohandle(mkarr(x, (nx,)), mkarr(y, (ny,)), d, udata)
        if 'exit' in d:
            runinfo['status'] = d
        return int(r) if r is not None else 0

    path = find_uopt()
    try:
        dll = CDLL(path)
    except BaseException as E:
        print("Failed to find libuopt", E)
        raise(E)

    cusolve = dll.usolve

    opts = fix_opts(opts)

    names  = [bytes(k, 'utf-8') for k in opts.keys()]
    values = [float(k) for k in opts.values()]

    cusolve.return_type = c_int
    cusolve.argtypes = (DPTR, DPTR, c_int64, DPTR, c_int64,
                        FFUNC, JFUNC, JVFUNC, JVFUNC, c_void_p, OFUNC,
                        CPTRPTR, DPTR, c_int64, c_void_p)
    cusolve.errcheck = my_errcheck

    ffunc  = FFUNC(fstub)
    Jfunc  = JFUNC(Jacstub if Jhandle is not None else 0)
    if Jvhandle is None:
        Jvfunc = JVFUNC(0)
    else:
        Jvfunc = JVFUNC(Jacvstub)
    JTvfunc  = JVFUNC(JacTvstub if JTvhandle is not None else 0)
    ofunc  = OFUNC(ostub)

    cx0 = np.ctypeslib.as_ctypes(x0)
    cxs = np.ctypeslib.as_ctypes(xs)
    cys = np.ctypeslib.as_ctypes(ys)
    STRARR = c_char_p * len(names)
    DBLARR = c_double * len(names)
    cnames = STRARR()
    cvalues = DBLARR()
    cnames[:] = names
    cvalues[:] = values

    res = cusolve(cx0, cxs, len(x0), cys, len(ys),
                  ffunc, Jfunc, Jvfunc, JTvfunc, None, ofunc,
                  cnames, cvalues, len(cnames), c_void_p())

    return (res, runinfo)

def my_errcheck(result, func, args):
    #print('my_errcheck %s: %d' %(func.__name__, result))
    return result
